/**
 * タブの表示制御を行なうクラス
 */
define(["dijit/registry", "dojo/_base/array", "dojo/_base/lang", "dojo/aspect", "dojo/cookie", "dojo/_base/declare", "dojo/dom", "dojo/_base/lang", "dojo/dom-construct", "dojo/dom-style", "dojo/query", "dojo/string", "dojo/topic", "dojo/has", "dojo/sniff"],
function(registry, array, lang, aspect, cookie, declare, dom, lang, domConstruct, domStyle, query, string, topic, has) {
  return declare(null, {
    constructor: function(tabSettings) {
      this.COOKIE_PREFIX = "__jfc_cookie_TabContainer_selectedTab_";
      this.defaultTabIds = tabSettings.defaultTabIds;
      this.enableSelectChild = tabSettings.enableSelectChild;
      // タブコンテナ毎にデフォルトで表示するタブを選択する関数を保持する配列
      this.selectTabFunc = {};
      this.initialTabs = {};
      // デフォルト表示タブの動作条件を設定する
      this.setSelectTabSettings(tabSettings);
      // 選択したタブを cookie に保存するイベントを設定する
      this.setPreserveSelectedTabEvent(tabSettings);
      // タブの表示・非表示の動作条件を設定する
      this.setTabVisibilitySettings(tabSettings);
    },

    init: function() {
      // デフォルトで表示するタブを選択する
      for (var key in this.selectTabFunc) {
        (this.selectTabFunc[key])();
      }
    },

    setSelectTabSettings: function(tabSettings) {
      if (!tabSettings.selectTabConditions) {
        return;
      }
      // タブコンテナ毎にデフォルトタブを選択する関数を生成する
      for (var tabContainerId in tabSettings.selectTabConditions) {
        var selectTabFunc = this.getSelectTabFunc(tabContainerId, tabSettings.selectTabConditions[tabContainerId]);
        if (selectTabFunc) {
          this.selectTabFunc[tabContainerId] = selectTabFunc;
        }
      }
    },

    getSelectTabFunc: function(tabContainerId, selectTabCondition) {
      var that = this;
      return function() {
        var tabContainer = registry.byId(tabContainerId);
        if (!tabContainer) {
          return;
        }
        var selectTabId = that.getConditionFilledTabId(tabContainerId, selectTabCondition);
        if (!selectTabId) {
          // 選択されていたタブをクッキーから取得
          selectTabId = cookie(that.COOKIE_PREFIX + tabContainerId);
        }
        if (!selectTabId) {
          // デフォルト選択のタブIDを返す
          selectTabId = that.defaultTabIds[tabContainerId];
        }
        // タブが存在するか確認後に選択する
        if (selectTabId && tabContainer.getIndexOfChild(registry.byId(selectTabId)) !== -1) {
          tabContainer.selectChild(selectTabId);
        }
      }
    },

    getConditionFilledTabId: function(tabContainerId, selectTabCondition) {
      // 選択条件を満たしたタブIDを返す
      if (selectTabCondition) {
        for (var ctrlId in selectTabCondition) {
          var selectTabIds =
            this.getConditionFilledTabIds(this.getCtrls(ctrlId), selectTabCondition[ctrlId]);
          if (Object.keys(selectTabIds).length === 0) {
            continue;
          }
          for (var tabId in selectTabIds) {
            return tabId;
          }
        }
      }
      return null;
    },

    getCtrlIds: function(ctrlId) {
      if (ctrlId.indexOf("/") === -1) {
        return [ctrlId];
      }
      ctrlId = ctrlId.replace("_002f", "\\_002f");
      ctrlId = ctrlId.replace("_005b-?w+_005d_002f_0040id", "\\_005b\\-?\\w*\\_005d\\_002f\\_0040id");
      var ctrlIdRegExp = ctrlId.split("/");
      // ctrlId の正規表現から widget を取得する
      var ctrlIdRe = new RegExp(ctrlIdRegExp[1]);
      var widgets = getAllWidgets();
      var widgets = array.filter(widgets, function(widget) {
        return ctrlIdRe.test(widget.id);
      });
      if (widgets && widgets.length > 0) {
        var widgetIds = [];
        array.forEach(widgets, function(widget) {
          widgetIds.push(widget.id);
        });
        return widgetIds;
      }
      var elements = array.filter(query("input,span", dojo.byId("centerarea")), function(input) {
        var id = input.id;
        if (!id || id === "") {
          return false;
        }
        return ctrlIdRe.test(id);
      });
      if (elements && elements.length > 0) {
        var elementIds = [];
        array.forEach(elements, function(element) {
          elementIds.push(element.id);
        });
        return elementIds;
      }
      return [];
    },

    getConditionFilledTabIds: function(ctrls, conditions) {
      var conditionFilledTabIds = {};
      for (var i=0,k=ctrls.length; i<k; i++) {
        var ctrlValue = this.getCtrlValue(ctrls[i]);
        for (var conditionValue in conditions) {
          var tabIds = conditions[conditionValue];
          if (ctrlValue !== conditionValue) {
            continue;
          }
          var tabIds = conditions[conditionValue];
          if (Object.prototype.toString.call(tabIds) === '[object Array]') {
            for (var j=0,l=tabIds.length; j<l; j++) {
              conditionFilledTabIds[tabIds[j]] = true;
            }
          } else {
            conditionFilledTabIds[tabIds] = true;
          }
        }
      }
      return conditionFilledTabIds;
    },

    getCtrlValue: function(ctrl) {
      // Dojo のウィジェット
      if (ctrl.baseClass) {
        if (ctrl.baseClass.lastIndexOf("dijitRadio", 0) === 0
            || ctrl.baseClass.lastIndexOf("dijitCheckBox", 0) === 0) {
          if (ctrl.get("checked")) {
            return ctrl.get("value") + "";
          } else {
            return null;
          }
        } else {
          return ctrl.get("value") + "";
        }
      }
      // input タグ
      if (ctrl.value) {
        return ctrl.value;
      }
      // 詳細表示、読み込み専用の span タグ
      if (ctrl.id && ctrl.id.lastIndexOf("content_", 0) === 0 ||
            ctrl.innerHTML) {
        return string.trim(ctrl.innerHTML);
      }
      return null;
    },

    getCtrls: function(ctrlId) {
      var ctrlIds = this.getCtrlIds(ctrlId);
      var ctrls = [];
      for (var i=0,l=ctrlIds.length; i<l; i++) {
        var ctrl = this.getCtrl(ctrlIds[i]);
        if (!ctrl) {
          continue;
        }
        ctrls.push(ctrl);
      }
      return ctrls;
    },

    getCtrl: function(id) {
      // Dojo のウィジェット
      var widget = registry.byId(id);
      if (widget) {
        return widget;
      }
      // input タグ
      var element = dom.byId(id);
      if (element) {
        return element;
      }
      // 詳細表示、読み込み専用の span タグ
      var span = dom.byId("content_" + id);
      if (span) {
        return span;
      }
      return null;
    },

    setTabVisibilitySettings: function(tabSettings) {
      if (!tabSettings.visibilityConditions) {
        return;
      }
      for (var tabContainerId in tabSettings.visibilityConditions) {
        // タブコンテナ毎にタブ表示の動作条件を設定する
        this.setTabVisibilitySetting(tabContainerId, tabSettings.visibilityConditions[tabContainerId]);
      }
    },

    saveInitialTabs: function(tabContainer) {
      var tabs = tabContainer.getChildren();
      tabs.forEach(function(tab, index) {
        tab.tabIndex = index;
      });
      this.initialTabs[tabContainer.id] = tabs;
    },

    // タブの表示条件の設定
    setTabVisibilitySetting: function(tabContainerId, visibilityConditions) {
      var tabContainer = registry.byId(tabContainerId);
      if (!tabContainer) {
        console.warn("tabContainerId:["+tabContainerId+"] not exists.");
        return;
      }
      // タブコンテナ毎に初期状態のタブを保存しておく
      this.saveInitialTabs(tabContainer);
      var srcCtrlsInCtrlId = {};
      for (ctrlId in visibilityConditions) {
        var conditions = visibilityConditions[ctrlId];
        // 初期表示時に設定されたタブを非表示にする
        for (var conditionValue in conditions) {
          this.invisibleTab(tabContainer, conditions[conditionValue]);
        }
        srcCtrlsInCtrlId[ctrlId] = this.getCtrls(ctrlId);
      }
      var tabVisiblityChangeFunc = this.getTabVisiblityChangeFunc(tabContainer, visibilityConditions, srcCtrlsInCtrlId);
      var isSetXHREvent = false;
      for (ctrlId in visibilityConditions) {
        var srcCtrls = srcCtrlsInCtrlId[ctrlId];
        for (var i=0,l=srcCtrls.length; i<l; i++) {
          if (srcCtrls[i].baseClass) {
            this.setTabVisiblityChangeFuncWidget(srcCtrls[i], tabVisiblityChangeFunc);
          } else if (!isSetXHREvent && srcCtrls[i].id && srcCtrls[i].id.lastIndexOf("content_", 0) === 0) {
            // タブの表示制御元が Dojo の Widget でない場合は、 XHR 実行後に設定を行なう
            this.setTabVisibilityChangeFuncAfterXHR(srcCtrls[i], tabVisiblityChangeFunc);
            isSetXHREvent = true;
          }
        }
      }
      // 初期表示用に即時実行する
      (tabVisiblityChangeFunc)();
    },

    getTabVisiblityChangeFunc: function(tabContainer, visibilityConditions, srcCtrls) {
      var that = this;
      return function() {
        var visibleTabIds = {};
        for (ctrlId in visibilityConditions) {
          var ctrls = srcCtrls[ctrlId];
          var conditions = visibilityConditions[ctrlId];
          lang.mixin(visibleTabIds, that.getConditionFilledTabIds(ctrls, conditions));
          for (var conditionValue in conditions) {
            var tabIds = conditions[conditionValue];
            for (var i=0,l=tabIds.length; i<l; i++) {
              var tabId = tabIds[i];
              if (visibleTabIds[tabId]) {
                that.visibleTab(tabContainer, tabId);
              } else {
                that.invisibleTab(tabContainer, tabId);
              }
            }
          }
        }
        that.changeTabContainerVisibility(tabContainer);
        // デフォルトで表示するタブを選択する
        (that.selectTabFunc[tabContainer.id])();
      }
    },

    setTabVisiblityChangeFuncWidget: function(widget, tabVisiblityChangeFunc) {
      var baseClass = widget.baseClass;
      if (!baseClass) {
        return;
      }
      var eventName;
      if (baseClass.lastIndexOf("dijitTextBox", 0) === 0 || baseClass.lastIndexOf("dijitSelect", 0) === 0) {
        if (widget.type === "hidden" || baseClass.lastIndexOf("dijitSelect", 0) === 0) {
          // dijit/form/Textbox の hidden 項目、dijit/form/Select の場合
          eventName = "change";
        } else {
          // 参照連動等で値が変更された場合に blur イベントが呼ばれるため、
          // keyup と blur にイベント登録しておく
          eventName = "keyup,blur";
        }
      } else if (baseClass.lastIndexOf("dijitRadio", 0) === 0 || baseClass.lastIndexOf("dijitCheckBox", 0) === 0) {
        eventName = "change";
      }
      this.setTabVisibilityChangeEvent(widget, eventName, tabVisiblityChangeFunc);

      if (!(baseClass.lastIndexOf("dijitRadio", 0) === 0 || baseClass.lastIndexOf("dijitCheckBox", 0) === 0)) {
        return;
      }
      // ラジオボタン、チェックボックスにクリアボタンがあればイベントを追加
      var clearButton = this.getClearButton(widget);
      if (!clearButton) {
        return;
      }
      clearButton.on("click", function() {
        tabVisiblityChangeFunc();
      });
    },

    setTabVisibilityChangeEvent: function(widget, eventName, tabVisiblityChangeFunc) {
      widget.on(eventName, function() {
        tabVisiblityChangeFunc();
      });
    },

    setTabVisibilityChangeFuncAfterXHR: function(widget, tabVisiblityChangeFunc) {
      aspect.after(window, "setResponseXMLContent", function() {
        tabVisiblityChangeFunc();
      });
    },

    getClearButton: function(widget) {
      return registry.byId("btnClear_" + widget.name);
    },

    getTab: function(tabContainerId, tabId) {
      var tabs = this.initialTabs[tabContainerId];
      for (var i=0,l=tabs.length; i<l; i++) {
        if (tabs[i].id === tabId) {
          tabs[i].tabIndex = i;
          return tabs[i];
        }
      }
      return null;
    },

    isTabVisible: function(tabContainer, tab) {
      if (tabContainer.getIndexOfChild(tab) === -1) {
        return false;
      }
      return true;
    },

    visibleTab: function(tabContainer, tabId) {
      var tab = this.getTab(tabContainer.id, tabId);
      if (this.isTabVisible(tabContainer, tab)) {
        return;
      }
      // タブコンテナが非表示状態でタブを追加するとタブの耳が表示されないので、
      // タブの追加前にタブコンテナを表示状態に変更する
      var styleDisplay = domStyle.get(tabContainer.domNode.parentNode, "display");
      if (styleDisplay && styleDisplay.indexOf("none") !== -1) {
        this.setTabVisibility(tabContainer, "block");
      }
      // hidden 項目に入力可・不可でセットされた disabled 設定を保存しておく
      var disabledValues = this.preserveDisabledProperty();
      // jfc_hidden_item 内へ退避しておいた hidden 項目を削除する
      this.removeHiddenNodes(tab);
      // 追加するタブの位置を取得する
      var tabIndex = this.getTabIndexForAddedTab(tabContainer, tab);
      try {
        tabContainer.addChild(tab, tabIndex);
      } catch (x) {
        console.error(x);
      }
      // hidden 項目に保存した disabled 設定を入力フォームに反映させる
      this.restoreDisabledProperty(tab, disabledValues);
    },

    getTabIndexForAddedTab: function(tabContainer, tab) {
      var tabIndex = 0;
      var tabs = this.initialTabs[tabContainer.id];
      if (tabs.length === 0) {
        return tabIndex;
      }
      if (tab.tabIndex > tabs.length) {
        return tabs.length;
      }
      for (var i=0,l=tabs.length; i<l; i++) {
        if (!this.isTabVisible(tabContainer, tabs[i])) {
          continue;
        }
        if (tab.tabIndex <= tabs[i].tabIndex) {
          break;
        }
        tabIndex++;
      }
      return tabIndex;
    },

    invisibleTab: function(tabContainer, tabId) {
      var tab = this.getTab(tabContainer.id, tabId);
      if (!this.isTabVisible(tabContainer, tab)) {
        return;
      }
      // 非表示になるタブの入力フィールドの値を hidden 項目として、
      // jfc_hidden_item 内へ保存しておく
      this.removeHiddenNodes(tab);
      this.moveInputNodesHidden(tab);
      if (tabContainer.getIndexOfChild(tab) !== -1) {
        tabContainer.removeChild(tab);
      }
    },

    setPreserveSelectedTabEvent: function() {
      if (!this.enableSelectChild) {
        return;
      }
      var that = this;
      // タブ選択時に id 値を cookie に保存するメソッドを追加する
      query(".dijitTabContainer").forEach(function(node, index, arr) {
        topic.subscribe(node.id + "-selectChild", function(child) {
          cookie(that.COOKIE_PREFIX + node.id, child.id);
        });
      });
    },

    preserveDisabledProperty: function() {
      var disabledValues = {};
      query("input[type='hidden']", dom.byId("jfc_hidden_item")).forEach(function(node) {
        var id = node.id || node.name;
        disabledValues[id] = node.disabled;
      });
      return disabledValues;
    },

    restoreDisabledProperty: function(tab, disabledValues) {
      // タブ内の入力フォームを取得する
      // 同じ id 値で disabled プロパティが設定されている場合、
      // フォームに反映させる
      var inputs = this.getInputNodes(tab);
      var radioGroupName = "";

      for (var i=0,l=inputs.length; i<l; i++) {
        var id = inputs[i].id || inputs[i].name;
        if (disabledValues[id] == undefined) {
          continue;
        }
        if (id.indexOf("btnClear_") === -1) {
          // ラジオボタンの場合、選択肢全てに disabled 属性を
          // セットするため、name 属性を用いる
          if (inputs[i].type === "radio") {
            if (radioGroupName === inputs[i].name) {
              continue;
            }
            radioGroupName = inputs[i].name;
            var radios = document.getElementsByName(radioGroupName);
            for (var j=0,m=radios.length; j<m; j++) {
              radios[j].disabled = disabledValues[id];
            }
          } else {
            inputs[i].disabled = disabledValues[id];
          }
        } else {
          dijit.byId(id).set("disabled", disabledValues[id]);
        }
      }
    },

    getInputNodes: function(tab) {
      var nodes = query("input[type='text']", tab.srcNodeRef);
      nodes = nodes.concat(query("input[type='checkbox']", tab.srcNodeRef));
      nodes = nodes.concat(query("input[type='radio']", tab.srcNodeRef));
      nodes = nodes.concat(query("input[type='button']", tab.srcNodeRef));
      // dijit/form/Select の値を保持している要素を取得する
      nodes = nodes.concat(query("input[data-dojo-attach-point='valueNode']", tab.srcNodeRef));
      nodes = nodes.concat(query("input[type='hidden']", tab.srcNodeRef));
      return nodes;
    },

    moveInputNodesHidden: function(tab) {
      var inputs = this.getInputNodes(tab);
      for (var i=0,l=inputs.length; i<l; i++) {
        var id = inputs[i].id || inputs[i].name;
        if (!id) {
          continue;
        }
        var hiddenItem = domConstruct.create("input", {
          type: "hidden",
          id: id,
          name: inputs[i].name || id,
        }, "jfc_hidden_item");

        // チェックボックス、ラジオボタンの場合には、
        // チェック時のみ value 値を保存する
        if (inputs[i].type === "radio" || inputs[i].type === "checkbox") {
          if (inputs[i].checked) {
            hiddenItem.value = inputs[i].value;
          }
        } else {
          hiddenItem.value = inputs[i].value;
        }
      }
    },

    removeHiddenNodes: function(tab) {
      var inputs = this.getInputNodes(tab);
      var hiddenInputs = {};
      for (var i=0,l=inputs.length; i<l; i++) {
        var id = inputs[i].id || inputs[i].name;
        if (!id || id === "") {
          continue;
        }
        hiddenInputs[id] = true;
      }
      query("input[type='hidden']", dom.byId("jfc_hidden_item")).forEach(function(node) {
        var id = node.id || node.name;
        if (id && hiddenInputs[id]) {
          domConstruct.destroy(id);
        }
      });
    },

    changeTabContainerVisibility: function(tabContainer) {
      var tabs = tabContainer.getChildren();
      if (tabs.length === 0) {
        // タブが存在しない場合、タブコンテナ自体を非表示にする
        this.setTabVisibility(tabContainer, "none");
      } else {
        this.setTabVisibility(tabContainer, "block");
      }
    },

    setTabVisibility: function(tabContainer, value) {
      this.setOpacity(tabContainer, value);
      domStyle.set(tabContainer.domNode.parentNode, "display", value);
      if (!has("ie")) {
        return;
      }
      // IE の場合、上記のみではタブコンテナの枠が消えないため、個別に表示を切り替える
      query("div[class*='dijitTabSpacer']", tabContainer.domNode).forEach(function(node) {
        domStyle.set(node, "display", value);
      });
      query("div[class*='dijitTabContainerTop-container']", tabContainer.domNode).forEach(function(node) {
        domStyle.set(node, "display", value);
      });
    },

    setOpacity: function(tabContainer, value) {
      if (!tabContainer.domNode.parentNode.style.opacity) {
        return;
      }
      var opacity = 0;
      if (value === "block") {
        opacity = 1;
      }
      domStyle.set(tabContainer.domNode.parentNode, "opacity", opacity);
    },

  });
});
